#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

HTMLChunker.py

Copyright 2009, 2014 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Splits a HTML file into chunks.

"""

from __future__ import unicode_literals


import os
import re
import copy

from lxml import etree
from six.moves import urllib

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import NS
from libgutenberg.Logger import debug, error, info

from ebookmaker.CommonCode import Options
options = Options()

MAX_CHUNK_SIZE = 100 * 1024  # bytes
MIN_CHUNK_SIZE = 1024  # bytes

SECTIONS = [
    ('div.section', 0.0),
    ('div.chapter', 0.0),
    ('section', 0.0),
    ('h1', 0.5),
    ('div', 0.5),
    ('h2', 0.7),
    ('h3', 0.75),
    ('p', 0.8),
    ('figure', 0.8),
]

NEVER_SPLIT = [NS.xhtml[tag] for tag in ['table', 'figure', 'dl', 'ol', 'ul']]

def xpath(node, path):
    """ xpath helper """
    return node.xpath(path, namespaces = gg.NSMAP)

def normalize_uri(uri):
    """ Normalize URI for idmap. """
    return urllib.parse.unquote(uri) # .decode('utf-8')


class HTMLChunker(object):
    """ Splits HTML tree into smaller chunks.

    Some epub viewers are limited in that they cannot display files
    larger than 300K.  If our HTML happens to be longer, we have to
    split it up.  Also smaller chunks do improve page flip times.


    """

    def __init__(self, version='epub2'):
        self.chunks = []
        self.idmap = {}
        self.chunk = None
        self.chunk_body = None
        self.chunk_size = 0
        self.next_id = 0
        self.version = version
        self.max_chunk_size = MAX_CHUNK_SIZE * (1 if version == 'epub2' else 3)
        self.nosplit = False

        self.tags = {}
        for tag, size in SECTIONS:
            self.tags[NS.xhtml[tag]] = max(MIN_CHUNK_SIZE, int(size * self.max_chunk_size))
        for tag in options.section_tags:
            self.tags[NS.xhtml[tag]] = MIN_CHUNK_SIZE


    def _make_name(self, url):
        """ Generate a name for the chunk. """
        u = list(urllib.parse.urlparse(url))
        root, ext = os.path.splitext(u[2])
        html_ext = 'html' if self.version == 'epub2' else 'xhtml'
        u[2] = f'{root}-{self.next_id}{ext}.{html_ext}'
        self.next_id += 1
        return urllib.parse.urlunparse(u)


    @staticmethod
    def make_template(tree):
        """ Make a copy with an empty html:body.

        This makes a template into which we can paste our chunks.

        """

        template = copy.deepcopy(tree)

        for c in xpath(template, '//xhtml:body'):

            # descend while elem has only one child
            while len(c) == 1:
                c = c[0]

            # clear children but save attributes
            attributes = c.attrib.items()
            c.clear()
            # was tentative fix for patological one-element-html case
            # for child in c:
            #     c.remove(child)
            for a in attributes:
                c.set(a[0], a[1])

        # debug(etree.tostring(template))

        return template


    def reset_chunk(self, template):
        """ start a new chunk """

        self.chunk = copy.deepcopy(template)
        self.chunk_size = 0
        self.chunk_body = xpath(self.chunk, "//xhtml:body")[0]
        while len(self.chunk_body) == 1:
            self.chunk_body = self.chunk_body[0]
        self.nosplit = False


    def shipout_chunk(self, attribs, chunk_id = None, comment = None):
        """ ready chunk to be shipped """

        attribs = copy.copy(attribs)

        if self.chunk_size > self.max_chunk_size and not self.nosplit:
            self.split(self.chunk, attribs)
            return

        url = normalize_uri(attribs.url)
        chunk_name = self._make_name(url)

        # the url of the whole page
        if url not in self.idmap:
            self.idmap[url] = chunk_name

        # fragments of the page
        for e in xpath(self.chunk, '//xhtml:*[@id]'):
            id_ = e.attrib['id']
            old_id = "%s#%s" % (url, id_)
            # key is unicode string,
            # value is uri-escaped byte string
            # if ids get cloned while chunking, map to the first one only
            if old_id not in self.idmap:
                self.idmap[old_id] = "%s#%s" % (
                    chunk_name,  urllib.parse.quote(id_))

        attribs.url = chunk_name
        attribs.id = chunk_id
        attribs.comment = comment
        self.chunks.append((self.chunk, attribs) )

        debug("Adding chunk %s (%d bytes) %s" % (chunk_name, self.chunk_size, chunk_id))


    def split(self, tree, attribs):
        """ Split whole html or split chunk.

        Find some arbitrary points to do it.

        """

        for body in xpath(tree, "//xhtml:body"):
            # we can't split a node that has only one child
            # descend while elem has only one child
            while len(body) == 1:
                body = body[0]

            debug("body tag is %s" % body.tag)

            template = self.make_template(tree)
            self.reset_chunk(template)

            # FIXME: is this ok ???
            # fixes patological one-element-body case
            self.chunk_body.text = body.text

            for child in body:
                if not isinstance(child, etree.ElementBase):
                    # comments, processing instructions etc.
                    continue

                # size measurement doesn't need to be exact
                try:
                    child_size = len(etree.tostring(child, encoding='utf-8'))
                except etree.SerialisationError:
                    child_size = len(etree.tostring(child, encoding='latin_1'))

                try:
                    tags = [child.tag + '.' + c for c in child.attrib['class'].split()]
                    tags.append(child.tag)
                except KeyError:
                    tags = [child.tag]

                for tag in tags:
                    if child.tag in NEVER_SPLIT:
                        self.nosplit = True
                        break
                    if ((self.chunk_size + child_size > self.max_chunk_size) or
                              (tag in self.tags and self.chunk_size > self.tags[tag])):

                        comment = ("Chunk: size=%d Split on %s"
                                   % (self.chunk_size, re.sub('^{.*}', '', tag)))
                        debug(comment)

                        # find a suitable id
                        chunk_id = None
                        for c in self.chunk_body:
                            if 'id' in c.attrib:
                                chunk_id = c.get('id')
                                break
                        debug("chunk id is: %s" % (chunk_id or ''))

                        self.shipout_chunk(attribs, chunk_id, comment)
                        self.reset_chunk(template)
                        break

                self.chunk_body.append(child)
                self.chunk_size = self.chunk_size + child_size

            # fixes patological one-element-body case
            self.chunk_body.tail = body.tail

            chunk_id = None
            if len(self.chunk_body):
                chunk_id = self.chunk_body[0].get('id')
            comment = "Chunk: size=%d" % self.chunk_size
            self.shipout_chunk(attribs, chunk_id, comment)
            self.reset_chunk(template)


    def rewrite_links(self, f):
        """ Rewrite all href and src using f(). """

        for chunk in self.chunks:
            # chunk['name'] = f(chunk['name'])

            for link in xpath(chunk[0], '//xhtml:*[@href]'):
                link.set('href', f(link.get('href')))

            for image in xpath(chunk[0], '//xhtml:*[@src]'):
                image.set('src', f(image.get('src')))

        for k, v in self.idmap.items():
            self.idmap[k] = f(v)


    def rewrite_internal_links(self):
        """ Rewrite links to point into right chunks.

        Because we split the HTML into chunks, all internal links need
        to be rewritten to become links into the right chunk.
        Rewrite all internal links in all chunks.

        """
        for chunk in self.chunks:
            for a in xpath(chunk[0], "//xhtml:*[@href]"):
                try:
                    uri = normalize_uri(a.get('href'))
                    a.set('href', self.idmap[uri])
                except KeyError:
                    ur, dummy_frag = urllib.parse.urldefrag(uri)
                    if ur in self.idmap:
                        error("HTMLChunker: Cannot rewrite internal link '%s'", uri)


    def rewrite_internal_links_toc(self, toc):
        """ Rewrite links to point into right chunks.

        Because we split the HTML into chunks, all internal links need
        to be rewritten to become links into the right chunk.
        Rewrite all links in the passed toc.

        """

        for entry in toc:
            try:
                entry[0] = self.idmap [normalize_uri(entry[0])]
            except KeyError:
                error("HTMLChunker: Cannot rewrite toc entry '%s'" % entry[0])
                error(repr(self.idmap))
                del entry



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

gutenberg.py

Copyright 2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Transforms for the Project Gutenberg flavor.

"""

from __future__ import unicode_literals

import datetime
import textwrap

from docutils import nodes
import docutils.transforms
import docutils.transforms.parts

from libgutenberg.Logger import error, warning, info, debug
from libgutenberg.DublinCore import DublinCore
from libgutenberg.GutenbergGlobals import PG_URL
from ebookmaker.mydocutils import nodes as mynodes

# pylint: disable=W0142

class SubRefToVarTransform (docutils.transforms.Transform):
    """
    Transforms subref nodes in 'pg' namespace into var nodes.

    We need to save some subrefs for later processing. The standard
    subref processing happens too early (ie. before docinfo is
    collected). So we transform subrefs into variables, await docinfo
    to be processed, and then process the variables.

    """

    default_priority = 219
    """ Before substitition def variables """


    def apply (self):
        for ref in self.document.traverse (nodes.substitution_reference):
            refname = ref['refname']
            if refname.startswith ('pg.'):
                var = mynodes.variable ()
                var['name'] = refname
                ref.replace_self (var)


class VariablesTransform (docutils.transforms.Transform):
    """ Replaces mynodes.var with parameters from metadata. """

    default_priority = 342
    """ After DocInfoCollector. """

    def apply(self):
        doc = self.document
        meta = doc.meta_block
        defs = doc.substitution_defs

        def getone (name, default = None):
            """ Get first value. """
            if name in meta:
                return meta[name][0]
            return default

        def getmany (name, default = []):
            """ Get list of all values. """
            return meta.get (name, default)

        def sub (var, nodes):
            var.replace_self (nodes)

        title = getone ('DC.Title', 'No Title')
        short_title = getone ('PG.Title', title)
        short_title = short_title.split ('\n', 1)[0]

        language = getmany ('DC.Language', ['en'])
        language = [DublinCore.language_map.get (
            x, default='Unknown').title () for x in language]
        language = DublinCore.strunk (language)

        copyrighted = getone ('PG.Rights', '').lower () == 'copyrighted'

        for variable in doc.traverse (mynodes.variable):
            name = variable['name']

            if name == 'pg.upcase-title':
                sub (variable, [ nodes.inline ('', short_title.upper ()) ])

            elif name == 'pg.produced-by':
                producers = getmany ('PG.Producer')
                if producers:
                    sub (variable, [ nodes.inline ('', 'Produced by %s.' %
                                                   DublinCore.strunk (producers)) ])
                else:
                    sub (variable, [])

            elif name == 'pg.credits':
                sub (variable, [ nodes.inline ('', getone ('PG.Credits', '')) ])

            elif name == 'pg.bibrec-url':
                url =  '%sebooks/%s' % (PG_URL, getone ('PG.Id', '999999'))
                sub (variable, [ nodes.reference ('', '', nodes.inline ('', url), refuri = url) ])

            elif name in ('pg.copyrighted-header', 'pg.copyrighted-footer'):
                if copyrighted:
                    subdef_copy = defs[name].deepcopy ()
                    sub (variable, subdef_copy.children)
                else:
                    sub (variable, [])

            elif name == 'pg.machine-header':
                tw = textwrap.TextWrapper (
                    width = 72,
                    initial_indent = 'Title: ',
                    subsequent_indent = ' ' * 7)

                if '\n' in title:
                    maintitle, subtitle = title.split ('\n', 1)
                    s = tw.fill (maintitle)
                    s += '\n'
                    tw.initial_indent = tw.subsequent_indent
                    s += tw.fill (subtitle)
                else:
                    s = tw.fill (title)
                s += '\n\n'

                tw.initial_indent = 'Author: '
                tw.subsequent_indent = ' ' * 8
                s += tw.fill (DublinCore.strunk (getmany ('DC.Creator', ['Unknown'])))
                s += '\n\n'

                date = getone ('PG.Released', '')
                try:
                    date = datetime.datetime.strptime (date, '%Y-%m-%d')
                    date = datetime.datetime.strftime (date, '%B %d, %Y')
                except ValueError:
                    date = 'unknown date'
                s += 'Release Date: %s [EBook #%s]\n' % (date, getone ('PG.Id', '999999'))

                for item in getmany ('PG.Reposted', []):
                    try:
                        date, comment = item.split (None, 1)
                    except ValueError:
                        date = item
                        comment = None
                    try:
                        date = datetime.datetime.strptime (date, '%Y-%m-%d')
                        date = datetime.datetime.strftime (date, '%B %d, %Y')
                    except ValueError:
                        date = 'unknown date'

                    s += 'Reposted: %s' % date
                    if comment:
                        s += ' [%s]' % comment
                    s += '\n'

                s += '\nLanguage: %s\n\n' % language

                sub (variable, [ nodes.inline ('', nodes.Text (s)) ])



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

EbookMaker.py

Copyright 2009-2014 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Stand-alone application to build EPUB and others out of html or rst.

"""

import argparse
import configparser
import collections
import datetime
import hashlib
import logging
import os.path
import re
import sys

from six.moves import cPickle

from libgutenberg.GutenbergGlobals import SkipOutputFormat
from libgutenberg.DublinCore import PGDCObject
import libgutenberg.GutenbergGlobals as gg
from libgutenberg.Logger import critical, debug, info, warning, error, exception
from libgutenberg import Logger, DublinCore
from libgutenberg import MediaTypes as mt
from libgutenberg import Cover

from ebookmaker import parsers
from ebookmaker import ParserFactory
from ebookmaker import Spider
from ebookmaker import WriterFactory
from ebookmaker.packagers import PackagerFactory
from ebookmaker import CommonCode
from ebookmaker.CommonCode import Options, dir_from_url, find_candidates
from ebookmaker.Version import VERSION

options = Options()

# store paths for system utilities in CONFIG_FILES[0]
# store default command line args in [default_args] section of CONFIG_FILES[1]
CONFIG_FILES = ['/etc/ebookmaker.conf', os.path.expanduser('~/.ebookmaker')]

DEPENDENCIES = collections.OrderedDict((
    ('all', ('html', 'epub', 'epub3', 'kindle', 'kf8', 'pdf', 'txt', 'rst')),
    ('test', ('html', 'epub', 'epub3')),
    ('html', ('html.images', 'html.noimages')),
    ('epub', ('epub.images', 'epub.noimages')),
    ('epub3', ('epub3.images',)),
    ('kindle', ('kindle.images', 'kindle.noimages')),
    ('kf8', ('kf8.images',)),
    ('pdf', ('pdf.images', 'pdf.noimages')),
    ('txt', ('txt.utf-8', 'txt.iso-8859-1', 'txt.us-ascii')),
    ('rst', ('rst.gen', )),
    ('kindle.noimages', ('epub.noimages', )),
    ('kindle.images', ('epub.images', )),
    ('kf8.images', ('epub3.images', )),
    ('html.noimages', ('picsdir.noimages', )),
    ('html.images', ('picsdir.images', )),
    ('pdf.noimages', ('picsdir.noimages', )),
    ('pdf.images', ('picsdir.images', )),
    ('rst.gen', ('picsdir.images', )),
))

BUILD_ORDER = """
picsdir.images picsdir.noimages
rst.gen
txt.utf-8 txt.iso-8859-1 txt.us-ascii
html.images html.noimages
epub.noimages kindle.noimages pdf.noimages
epub.images kindle.images pdf.images
epub3.images kf8.images
cover.small cover.medium
qrcode rdf facebook twitter null""".split()

FILENAMES = {
    'html.noimages': '{id}-noimages-h.html',
    'html.images': '{id}-h.html',

    'epub.noimages': '{id}-epub.epub',
    'epub.images': '{id}-images-epub.epub',
    'epub3.images': '{id}-images-epub3.epub',

    'kindle.noimages': '{id}-kindle.mobi',
    'kindle.images': '{id}-images-kindle.mobi',
    'kf8.images': '{id}-kf8-kindle.mobi',

    'pdf.noimages': '{id}-pdf.pdf',
    'pdf.images': '{id}-images-pdf.pdf',

    'txt.utf-8': '{id}-0.txt',
    'txt.iso-8859-1': '{id}-8.txt',
    'txt.us-ascii': '{id}.txt',

    'rst.gen': '{id}-rst.rst',

    'picsdir.noimages': '{id}-noimages.picsdir',  # do we need this ?
    'picsdir.images': '{id}-images.picsdir',  # do we need this ?

    'cover': '{id}-cover.png'
}

COVERPAGE_MIN_AREA = 200 * 200

def make_output_filename(type_, dc):
    """ Make a suitable filename for output type. """
    name_token = options.outputfile or dc.project_gutenberg_id
    if name_token:
        # PG book: use PG naming convention
        return FILENAMES[type_].format(id=name_token)
    # not a PG ebook
    return FILENAMES[type_].format(id=gg.string_to_filename(dc.title)[:65])


def cover_file_filter(fpath):
    dirpath, fname = os.path.split(fpath)
    if 'cover' not in fname.lower():
        return False
    name, ext = os.path.splitext(fpath)
    if ext not in ['.jpg', '.jpeg', '.png', '.gif']:
        return False
    return os.access(fpath, os.R_OK)

def add_cover(cover_url, spider):
    cover_parser = ParserFactory.ParserFactory.create(cover_url)
    cover_parser.pre_parse()
    if check_cover_size(cover_parser):
        cover_parser.attribs.rel.add('icon')
        if cover_parser.attribs.url not in spider.parsed_urls:
            spider.parsers.append(cover_parser)
        return True
    return False

def check_cover_size(p):
    if hasattr(p, 'get_image_dimen'):
        dimen = p.get_image_dimen()
        if (dimen[0] * dimen[1]) > COVERPAGE_MIN_AREA:
            return True
        else:
            p_url = p.url if hasattr(p, 'url') else ''
            warning("coverpage candidate %s is too small (%d x %d)" %
                    (p_url, dimen[0], dimen[1]))
    return False


def elect_coverpage(spider, url, dc):
    """ Find first coverpage candidate that is not too small. """

    coverpage_found = False
    for p in spider.parsers:
        if 'icon' in p.attribs.rel:
            if hasattr(p, 'get_image_dimen'):
                if not check_cover_size(p):
                    p.attribs.rel.remove('icon')
                    continue
            if coverpage_found:
                # keep the first one found, reset all others
                p.attribs.rel.remove('icon')
                continue
            coverpage_found = True

    # check sourcedir for a cover by name
    if not coverpage_found:
        for cover_url in find_candidates(url, file_filter=cover_file_filter):
            if add_cover(cover_url, spider):
                coverpage_found = True
                break

    if spider.parsers and not coverpage_found and options.generate_cover:
        if not hasattr(Cover, 'cairo'):
            warning('Cairo not installed, cover generation disabled')
            return
        if options.outputdir:
            dir = options.outputdir
        else:
            dir = dir_from_url(url)
        debug('generating cover in %s' % dir)
        cover_url = generate_cover(dir, dc)
        if cover_url:
            add_cover(cover_url, spider)


def generate_cover(dir, dc):
    try:
        cover_image = Cover.draw(dc, cover_width=1200, cover_height=1800)
        cover_url = os.path.join(dir, make_output_filename('cover', dc))
        with open(cover_url, 'wb+') as cover:
            cover_image.save(cover)
        return cover_url
    except OSError:
        error("OSError, Cairo not installed or couldn't write file.")
        return None

def get_dc(job):
    """ Get DC for book. """
    url = job.url
    parser = ParserFactory.ParserFactory.create(url)
    parser.parse()
    if options.is_job_queue:
        dc = PGDCObject()
        dc.load_from_database(job.ebook)
        dc.source = job.source
        dc.opf_identifier = job.opf_identifier
        return dc

    # this is needed because the the document is not parsed again
    if options.coverpage_url:
        parser._make_coverpage_link(coverpage_url=options.coverpage_url)


    dc = DublinCore.GutenbergDublinCore()
    try:
        dc.load_from_rstheader(parser.unicode_content())
    except (ValueError, UnicodeError):
        debug("No RST header found.")
        try:
            dc.load_from_parser(parser)
        except (ValueError, AttributeError, UnicodeError):
            debug("No HTML header found.")
            try:
                dc.load_from_pgheader(parser.unicode_content())
            except (ValueError, UnicodeError):
                debug("No PG header found.")

    dc.source = parser.attribs.url
    dc.title = options.title or dc.title or 'NA'

    if options.author:
        dc.add_author(options.author, 'cre')
    if not dc.authors:
        dc.add_author('NA', 'cre')

    dc.project_gutenberg_id = options.ebook or dc.project_gutenberg_id
    if dc.project_gutenberg_id:
        dc.opf_identifier = ('%sebooks/%d' % (gg.PG_URL, dc.project_gutenberg_id))
    else:
        dc.opf_identifier = ('urn:mybooks:%s' %
                             hashlib.md5(dc.source.encode('utf-8')).hexdigest())

    # We need a language to build a valid epub, so just make one up.
    if not dc.languages:
        info('no language found, using default')
        dc.add_lang_id('en')
    return dc


def add_local_options(ap):
    """ Add local options to commandline. """

    ap.add_argument(
        '--version',
        action='version',
        version="%%(prog)s %s" % VERSION
    )

    ap.add_argument(
        "--make",
        dest="types",
        choices=CommonCode.add_dependencies(['all', 'test'], DEPENDENCIES),
        default=[],
        action='append',
        help="output type (default: all)")

    ap.add_argument(
        "--max-depth",
        metavar="LEVELS",
        dest="max_depth",
        type=int,
        default=1,
        help="go how many levels deep while recursively retrieving pages. "
        "(0 == infinite) (default: %(default)s)")

    ap.add_argument(
        "--strip_links",
        dest="strip_links",
        action="store_true",
        help="strip  <a href='external address' /> links")

    ap.add_argument(
        "--include",
        metavar="GLOB",
        dest="include_urls",
        default=[],
        action="append",
        help="include urls (repeat for more) (default: urls under the same directory)")

    ap.add_argument(
        "--exclude",
        metavar="GLOB",
        dest="exclude_urls",
        default=[],
        action="append",
        help="exclude urls from included urls (repeat for more) (default: none)")

    ap.add_argument(
        "--include-mediatype",
        metavar="GLOB/GLOB",
        dest="include_mediatypes",
        default=mt.TEXT_MEDIATYPES | mt.AUX_MEDIATYPES,
        action="append",
        help="include mediatypes (repeat for more) (eg. 'image/*') "
        "(default: most common text mediatypes)")

    ap.add_argument(
        "--exclude-mediatype",
        metavar="GLOB/GLOB",
        dest="exclude_mediatypes",
        default=[],
        action="append",
        help="exclude this mediatype from included mediatypes "
        "(repeat for more)")

    ap.add_argument(
        "--input-mediatype",
        metavar="MEDIATYPE",
        dest="input_mediatype",
        default=None,
        help="mediatype of input url (default: http response else file extension)")

    ap.add_argument(
        "--mediatype-from-extension",
        dest="mediatype_from_extension",
        action="store_true",
        default=False,
        help="guess all mediatypes from file extension, overrides http response")

    ap.add_argument(
        "--rewrite",
        metavar="from>to",
        dest="rewrite",
        default=[],
        action="append",
        help="rewrite url eg. 'http://www.example.org/>http://www.example.org/index.html'")

    ap.add_argument(
        "--title",
        dest="title",
        default=None,
        help="ebook title (default: from meta)")

    ap.add_argument(
        "--author",
        dest="author",
        default=None,
        help="author (default: from meta)")

    ap.add_argument(
        "--ebook",
        dest="ebook",
        type=int,
        default=0,
        help="ebook no. (default: from meta)")

    ap.add_argument(
        "--output-dir",
        metavar="OUTPUT_DIR",
        dest="outputdir",
        default="./",
        help="output directory (default: ./)")

    ap.add_argument(
        "--config-dir",
        metavar="CONFIG_DIR",
        dest="configdir",
        default="",
        help="config directory (default: ''")

    ap.add_argument(
        "--output-file",
        metavar="OUTPUT_FILE",
        dest="outputfile",
        default=None,
        help="token for use in filenames (default: <ebook number>)")

    ap.add_argument(
        "--section",
        metavar="TAG.CLASS",
        dest="section_tags",
        default=[],
        action="append",
        help="split epub on TAG.CLASS")

    ap.add_argument(
        "--packager",
        dest="packager",
        choices=['ww', 'gzip'],
        default=None,
        help="PG internal use only: which packager to use (default: none)")

    ap.add_argument(
        "--cover",
        dest="coverpage_url",
        default=None,
        help="use the cover specified by an absolute url")

    ap.add_argument(
        "--generate_cover",
        dest="generate_cover",
        action="store_true",
        help="if no cover is specified by the source, or as an argument, generate a cover")

    ap.add_argument(
        "--jobs",
        dest="is_job_queue",
        action="store_true",
        help="PG internal use only: read pickled job queue from stdin")

    ap.add_argument(
        "--extension-package",
        metavar="PYTHON_PACKAGE",
        dest="extension_packages",
        default=[],
        action="append",
        help="PG internal use only: load extensions from package")

    ap.add_argument(
        "url",
        help="url of file to convert")


def open_log(path):
    """ Open a logfile in the output directory. """
    if options.notify:
        Logger.notifier = CommonCode.queue_notifications
    file_handler = Logger.setup(
        Logger.LOGFORMAT,
        logfile=path,
        loglevel=logging.INFO,
    )
    return file_handler


def close_log(handler):
    """ Close logfile handler. """
    if handler:
        logging.getLogger().removeHandler(handler)
        handler.close()


def do_job(job):
    """ Do one job. """

    log_handler = None
    Logger.ebook = job.ebook
    if job.logfile:
        log_handler = open_log(os.path.join(os.path.abspath(job.outputdir), job.logfile))

    debug('=== Building %s ===' % job.type)
    start_time = datetime.datetime.now()
    try:
        if job.url:
            spider = Spider.Spider(job)

            for rewrite in options.rewrite:
                from_url, to_url = rewrite.split('>')
                spider.add_redirection(from_url, to_url)

            attribs = parsers.ParserAttributes()
            attribs.url = parsers.webify_url(job.url)
            attribs.id = 'start'


            if options.input_mediatype:
                attribs.orig_mediatype = attribs.HeaderElement.from_str(
                    options.input_mediatype)

            spider.recursive_parse(attribs)
            if job.type.split('.')[0] in ('epub', 'epub3', 'html', 'kindle', 'cover', 'pdf'):
                elect_coverpage(spider, job.url, job.dc)
            job.url = spider.redirect(job.url)
            job.base_url = job.url
            job.spider = spider

        writer = WriterFactory.create(job.maintype)
        writer.build(job)

        if options.validate:
            writer.validate(job)

        packager = PackagerFactory.create(options.packager, job.type)
        if packager:
            packager.package(job)

        if job.type == 'html.images':
            # FIXME: hack for push packager
            options.html_images_list = list(job.spider.aux_file_iter())

        if job.type.split('.')[0] == 'txt':
            # don't us GutenbergTextParser for subsequent builds
            ParserFactory.ParserFactory.parsers = {}

    except SkipOutputFormat as what:
        warning("%s" % what)

    except Exception as what:
        exception("%s" % what)

    end_time = datetime.datetime.now()
    info(' %s made in %s' % (job.type, end_time - start_time))

    if log_handler:
        close_log(log_handler)


def config():
    """ Process config files and commandline params. """

    ap = argparse.ArgumentParser(prog='EbookMaker')
    CommonCode.add_common_options(ap, CONFIG_FILES[1])
    add_local_options(ap)
    CommonCode.set_arg_defaults(ap, CONFIG_FILES[1])

    global options
    options.update(vars(CommonCode.parse_config_and_args(
        ap,
        CONFIG_FILES[0],
        {
            'proxies' : None,
            'xelatex' : 'xelatex',
            'mobigen' : 'ebook-convert',
            'mobilang': 'ebook-convert',
            'mobikf8': 'ebook-convert',
            'groff'   : 'groff',
            'rhyming_dict': None,
            'timestamp': datetime.datetime.today().isoformat()[:19],
        }
    )))

    if not re.search(r'^(https?|file):', options.url):
        options.url = os.path.abspath(options.url)

def main():
    """ Main program. """

    try:
        config()
    except configparser.Error as what:
        error("Error in configuration file: %s", str(what))
        return 1

    Logger.set_log_level(options.verbose)

    options.types = options.types or ['all']
    options.types = CommonCode.add_dependencies(options.types, DEPENDENCIES, BUILD_ORDER)
    debug("Building types: %s" % ' '.join(options.types))
    start_time = datetime.datetime.now()

    ParserFactory.load_parsers()
    WriterFactory.load_writers()
    PackagerFactory.load_packagers()

    output_files = dict()
    if options.is_job_queue:
        job_queue = cPickle.load(sys.stdin.buffer) # read bytes

    else:
        job_queue = []
        for type_ in options.types:
            job = CommonCode.Job(type_)
            job.url = options.url
            job.ebook = options.ebook
            job.outputdir = options.outputdir
            job_queue.append(job)

    dc = None
    for job in job_queue:
        try:
            info('Job starting for type %s from %s', job.type, job.url)
            dc = get_dc(job) # this is when doc at job.url gets parsed!
            job.outputfile = job.outputfile or make_output_filename(job.type, dc)
            output_files[job.type] = job.outputfile
            if job.type.startswith('kindle'):
                absoutputdir = os.path.abspath(job.outputdir)
                if job.type == 'kindle.images':
                    job.url = os.path.join(absoutputdir, output_files['epub.images'])
                elif job.type == 'kindle.noimages':
                    job.url = os.path.join(absoutputdir, output_files['epub.noimages'])
            if job.type.startswith('kf8'):
                absoutputdir = os.path.abspath(job.outputdir)
                job.url = os.path.join(absoutputdir, output_files['epub3.images'])

            options.outputdir = job.outputdir
            job.dc = dc
            do_job(job)
            if dc and hasattr(dc, 'session') and dc.session:
                dc.session.close()
                dc.session = None # probably overkill
        except Exception as e:
            error('Job failed for type %s from %s', job.type, job.url)
            exception(e)

    packager = PackagerFactory.create(options.packager, 'push')
    if packager:
        # HACK: the WWers ever only convert one ebook at a time
        job = job_queue[0]
        job.outputfile = '%d-final.zip' % (dc.project_gutenberg_id)
        packager.package(job)

    end_time = datetime.datetime.now()
    info(' Finished jobs. Total time: %s' % (end_time - start_time))
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""
RSTWriter.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Build an RST file. This is just the master RST with the PG license mixed in.

"""


import os

from libgutenberg.Logger import debug, info, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat
from ebookmaker import ParserFactory
from ebookmaker import writers

class Writer (writers.BaseWriter):
    """ Class to write a reStructuredText. """

    def build (self, job):
        """ Build RST file. """

        filename = os.path.join (os.path.abspath(job.outputdir), job.outputfile)

        info ("Creating RST file: %s" % filename)

        parser = ParserFactory.ParserFactory.create (job.url)

        if not hasattr (parser, 'rst2nroff'):
            error ('RSTWriter can only work on a RSTParser.')
            raise SkipOutputFormat

        data = parser.preprocess ('utf-8').encode ('utf-8')

        self.write_with_crlf (filename, data)

        info ("Done RST file: %s" % filename)



#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
Unitame.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Module to implement the totally superfluous PG plain text conversion
into long extinct encodings.

We have to unitame-translate before feeding to nroff because nroff
does some irreversible (and wrong) translations of its own, like ä ->
a. Also, some unitame-translations change the number of characters,
thus throwing already-justified text off.

We cannot do the translations before feeding the source to docutils
because if we change the length of titles, we get the warning: Title
underline too short.

Translation does some dangerous things, like converting quotes to
apostrophes, which are command escapes in nroff. We have to escape
apostrophes in the source text but not apostroph-commands inserted by
the converter.

We also have to translate some important non-ascii characters, like
nbsp and shy, into command sequences before they reach unitame because
unitame would convert them into the semantically different space and
hyhpen.

All this makes translation inside the docutils converter the best
choice. Implemented as a docutils translator that visits all text
nodes.

Smart quote translation should also go into a docutils
translator. Likewise a translator for text-transform: upper.

"""

from __future__ import unicode_literals

import codecs
import unicodedata as ud

# UnitameData is generated from unitame.dat
from ebookmaker.UnitameData import unicode_to_iso_8859_1, iso_8859_1_to_ascii

# tweak dicts for translate ()
u2i = dict ( [ (ord (o), s) for o, s in unicode_to_iso_8859_1.items () ] )
i2a = dict ( [ (ord (o), s) for o, s in iso_8859_1_to_ascii.items () ] )

u2i.update ( {
    0x2000:     ' ',    # en quad
    0x2001:     '  ',   # em quad
    0x2002:     ' ',    # en space
    0x2003:     '  ',   # em space
    0x2004:     ' ',    # 3/em space
    0x2005:     '',     # 4/em
    0x2006:     '',     # 6/em
    0x2007:     ' ',    # figure space
    0x2008:     '',     # punctuation space
    0x2009:     '',     # thin space
    0x200a:     '',     # hair space
    0x200b:     '',     # zero space
    0x200c:     '',     # zwnj
    0x200d:     '',     # zwj
    0x2010:     '-',    # hyphen
    0x2011:     '-',    # non-breaking hyphen
    0x2012:     '-',    # figure-dash
    0x2013:     '-',    # en dash
    0x2014:     '--',   # em dash
    0x2015:     '-',    # horizontal bar
    0x2026:     '...',  # horizontal ellipsis
    ord ('™'): '(tm)',
    ord ('‹'): '<',
    ord ('›'): '>',
    ord ('†'): '+',
    ord ('‡'): '++',
    ord ('⁑'): '**',
    ord ('⁂'): '***',
    ord ('•'): '-',
    ord ('′'): '´',
    ord ('″'): '´´',
    ord ('‴'): '´´´',
    ord ('⁗'): '´´´´',
    ord ('⁓'): '~',
    ord ('‰'): '%o',
    ord ('‱'): '%oo',
    ord ('⚹'): '*',    # U+26b9 sextile
    ord ('⁰'): '^0',
    ord ('⁴'): '^4',
    ord ('⁵'): '^5',
    ord ('⁶'): '^6',
    ord ('⁷'): '^7',
    ord ('⁸'): '^8',
    ord ('⁹'): '^9',
    } )

# somehow cram these into ascii, so the ppers stop whining about not
# having nbsp in ascii, then fix it later by replacing them with nroff
# commands.

i2a.update ( {
    ord ('¹'): '^1',
    ord ('²'): '^2',
    ord ('³'): '^3',
    0x00a0:     '\u0011',       # nbsp => DC1
    0x00ad:     '\u0012',       # shy  => DC2
} )

unhandled_chars = []

def strip_accents (text):
    """ Strip accents from string.

    If the accented character doesn't fit into the encoding,
    remove the accent and try again.

    """
    return ud.normalize (
        'NFKC',
        ''.join ([c for c in ud.normalize ('NFKD', text) if ud.category (c) != 'Mn'])
    )


def unitame (exc):
    """
    Encoding error handler.

    The encoder handles all compatible characters itself.  It calls
    this function whenever it encounters a character it cannot encode.
    This function searches the unitame database for a replacement.


    """

    l = []
    for cc in exc.object[exc.start:exc.end]:
        c = cc
        if exc.encoding == 'latin-1': # python name for iso-8859-1
            c = c.translate (u2i)
            c = strip_accents (c)
            if c and ord (max (c)) < 256:
                l.append (c)
                c = None
        elif exc.encoding == 'ascii': # python name for us-ascii
            # "1¼" -> "1 1/4"
            if cc in '¼½¾':
                if exc.start > 0 and exc.object[exc.start - 1] in '0123456789':
                    l.append (' ')
            c = c.translate (u2i)
            c = c.translate (i2a)
            c = strip_accents (c)
            if c and ord (max (c)) < 128:
                l.append (c)
                c = None

        if c:
            l.append ('{~%s U+%04x~}' % (ud.name (cc), ord (cc)))
            unhandled_chars.extend (l)

    return (''.join (l), exc.end)


codecs.register_error ('unitame', unitame)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
HtmlTemplates.py

Copyright 2022 by Project Gutenberg

Use f-strings to render boilerplate trees
"""
import datetime
import html
import lxml
from lxml import etree

pg_date = datetime.date(1971, 12, 1)
try:
    hr_format = "%B %-d, %Y"
    f'{pg_date.strftime(hr_format)}'
except ValueError:
    # https://strftime.org/
    hr_format = "%B %#d, %Y"


def pgheader(dc):
    def pstyle(key, val):
        if not val:
            return ''
        return f"<p style='display:block; margin-top:1em; margin-bottom:1em; margin-left:2em; text-indent:-2em'><strong>{key}</strong>: {html.escape(val)}</p>"

    language_list = []
    lang = ''
    for language in dc.languages:
        lang = lang if lang else language.id 
        language_list.append(language.language)

    if 'copyright' in dc.rights:
        rights = 'This is a *copyrighted* Project Gutenberg eBook, details below.'
    else:
        rights = '''
This ebook is for the use of anyone anywhere in the United States and most other parts of the world at no cost and with almost no restrictions whatsoever. You may copy it, give it away or re-use it under the terms of the <a class="reference internal" href="#project-gutenberg-license">Project Gutenberg License</a> included with this ebook or online at <a class="reference external" href="https://www.gutenberg.org/license">https://www.gutenberg.org/license</a>. If you are not located in the United States, you’ll have to check the laws of the country where you are located before using this eBook.'''

    pg_header = f'''
<section class="pg-boilerplate pgheader" id="pg-header" xml:lang="en" lang="en" xmlns="http://www.w3.org/1999/xhtml">
    <h2 style='text-align:center; font-size:1.2em; font-weight:bold'>The Project Gutenberg eBook of <span lang='{lang}' xml:lang='{lang}'>{html.escape(dc.title_no_subtitle)}</span>, by {html.escape(dc.authors_short())}</h2>
    <div style='display:block; margin:1em 0'>{rights}</div>

    <div class="container" id="pg-machine-header">
        {pstyle('Title', dc.title_no_subtitle)}
        {pstyle('Subtitle', dc.subtitle)}
        {pstyle('Author', dc.authors_short())}
        {pstyle('Release Date', 
            f'{dc.release_date.strftime(hr_format)} [EBook #{dc.project_gutenberg_id}]')}
        {pstyle('Language', ', '.join(language_list))}
        {pstyle('Original Publication', dc.pubinfo)}
        {pstyle('Credits', dc.credit)}
    </div>
    <div class="vspace" style="height: 2em"><br /></div>
        <div style="text-align:center">
            <span>*** START OF THIS PROJECT GUTENBERG EBOOK ***</span>
        </div>
</section>
'''
    return etree.fromstring(pg_header, lxml.html.XHTMLParser())
    

def pgfooter(dc):
    copyright_addition = '''<p>This particular work is one of the few individual works protected
by copyright law in the United States and most of the remainder of the
world, included in the Project Gutenberg collection with the
permission of the copyright holder. Information on the copyright owner
for this particular work and the terms of use imposed by the copyright
holder on this work are set forth at the beginning of this work.</p>
    ''' if 'copyright' in dc.rights else ''

    pg_footer = f'''
<section class="pg-boilerplate pgheader" id="pg-footer" lang='en' xml:lang='en' xmlns="http://www.w3.org/1999/xhtml">
        <div style="text-align:center">
            <span>*** END OF THIS PROJECT GUTENBERG EBOOK ***</span>
        </div>

<div style='display:block; margin:1em 0'>
Updated editions will replace the previous one&#8212;the old editions will
be renamed.
</div>

<div style='display:block; margin:1em 0'>
Creating the works from print editions not protected by U.S. copyright
law means that no one owns a United States copyright in these works,
so the Foundation (and you!) can copy and distribute it in the United
States without permission and without paying copyright
royalties. Special rules, set forth in the General Terms of Use part
of this license, apply to copying and distributing Project
Gutenberg&#8482; electronic works to protect the PROJECT GUTENBERG&#8482;
concept and trademark. Project Gutenberg is a registered trademark,
and may not be used if you charge for an eBook, except by following
the terms of the trademark license, including paying royalties for use
of the Project Gutenberg trademark. If you do not charge anything for
copies of this eBook, complying with the trademark license is very
easy. You may use this eBook for nearly any purpose such as creation
of derivative works, reports, performances and research. Project
Gutenberg eBooks may be modified and printed and given away&#8212;you may
do practically ANYTHING in the United States with eBooks not protected
by U.S. copyright law. Redistribution is subject to the trademark
license, especially commercial redistribution.
</div>

<div style='margin-top:1em; font-size:1.1em; text-align:center' id='project-gutenberg-license'>START: FULL LICENSE</div>
<h2 style='text-align:center;font-size:0.9em'>THE FULL PROJECT GUTENBERG LICENSE</h2>
<div style='text-align:center;font-size:0.9em'>PLEASE READ THIS BEFORE YOU DISTRIBUTE OR USE THIS WORK</div>

<div style='display:block; margin:1em 0'>
To protect the Project Gutenberg&#8482; mission of promoting the free
distribution of electronic works, by using or distributing this work
(or any other work associated in any way with the phrase &#8220;Project
Gutenberg&#8221;), you agree to comply with all the terms of the Full
Project Gutenberg&#8482; License available with this file or online at
www.gutenberg.org/license.
</div>

<div style='display:block; font-size:1.1em; margin:1em 0; font-weight:bold'>
Section 1. General Terms of Use and Redistributing Project Gutenberg&#8482; electronic works
</div>

<div style='display:block; margin:1em 0'>
1.A. By reading or using any part of this Project Gutenberg&#8482;
electronic work, you indicate that you have read, understand, agree to
and accept all the terms of this license and intellectual property
(trademark/copyright) agreement. If you do not agree to abide by all
the terms of this agreement, you must cease using and return or
destroy all copies of Project Gutenberg&#8482; electronic works in your
possession. If you paid a fee for obtaining a copy of or access to a
Project Gutenberg&#8482; electronic work and you do not agree to be bound
by the terms of this agreement, you may obtain a refund from the person
or entity to whom you paid the fee as set forth in paragraph 1.E.8.
</div>

<div style='display:block; margin:1em 0'>
1.B. &#8220;Project Gutenberg&#8221; is a registered trademark. It may only be
used on or associated in any way with an electronic work by people who
agree to be bound by the terms of this agreement. There are a few
things that you can do with most Project Gutenberg&#8482; electronic works
even without complying with the full terms of this agreement. See
paragraph 1.C below. There are a lot of things you can do with Project
Gutenberg&#8482; electronic works if you follow the terms of this
agreement and help preserve free future access to Project Gutenberg&#8482;
electronic works. See paragraph 1.E below.
</div>

<div style='display:block; margin:1em 0'>
1.C. The Project Gutenberg Literary Archive Foundation (&#8220;the
Foundation&#8221; or PGLAF), owns a compilation copyright in the collection
of Project Gutenberg&#8482; electronic works. Nearly all the individual
works in the collection are in the public domain in the United
States. If an individual work is unprotected by copyright law in the
United States and you are located in the United States, we do not
claim a right to prevent you from copying, distributing, performing,
displaying or creating derivative works based on the work as long as
all references to Project Gutenberg are removed. Of course, we hope
that you will support the Project Gutenberg&#8482; mission of promoting
free access to electronic works by freely sharing Project Gutenberg&#8482;
works in compliance with the terms of this agreement for keeping the
Project Gutenberg&#8482; name associated with the work. You can easily
comply with the terms of this agreement by keeping this work in the
same format with its attached full Project Gutenberg&#8482; License when
you share it without charge with others.
</div>

{copyright_addition}
<div style='display:block; margin:1em 0'>
1.D. The copyright laws of the place where you are located also govern
what you can do with this work. Copyright laws in most countries are
in a constant state of change. If you are outside the United States,
check the laws of your country in addition to the terms of this
agreement before downloading, copying, displaying, performing,
distributing or creating derivative works based on this work or any
other Project Gutenberg&#8482; work. The Foundation makes no
representations concerning the copyright status of any work in any
country other than the United States.
</div>

<div style='display:block; margin:1em 0'>
1.E. Unless you have removed all references to Project Gutenberg:
</div>

<div style='display:block; margin:1em 0'>
1.E.1. The following sentence, with active links to, or other
immediate access to, the full Project Gutenberg&#8482; License must appear
prominently whenever any copy of a Project Gutenberg&#8482; work (any work
on which the phrase &#8220;Project Gutenberg&#8221; appears, or with which the
phrase &#8220;Project Gutenberg&#8221; is associated) is accessed, displayed,
performed, viewed, copied or distributed:
</div>

<blockquote>
  <div style='display:block; margin:1em 0'>
    This eBook is for the use of anyone anywhere in the United States and most
    other parts of the world at no cost and with almost no restrictions
    whatsoever. You may copy it, give it away or re-use it under the terms
    of the Project Gutenberg License included with this eBook or online
    at <a href="https://www.gutenberg.org">www.gutenberg.org</a>. If you
    are not located in the United States, you will have to check the laws
    of the country where you are located before using this eBook.
  </div>
</blockquote>

<div style='display:block; margin:1em 0'>
1.E.2. If an individual Project Gutenberg&#8482; electronic work is
derived from texts not protected by U.S. copyright law (does not
contain a notice indicating that it is posted with permission of the
copyright holder), the work can be copied and distributed to anyone in
the United States without paying any fees or charges. If you are
redistributing or providing access to a work with the phrase &#8220;Project
Gutenberg&#8221; associated with or appearing on the work, you must comply
either with the requirements of paragraphs 1.E.1 through 1.E.7 or
obtain permission for the use of the work and the Project Gutenberg&#8482;
trademark as set forth in paragraphs 1.E.8 or 1.E.9.
</div>

<div style='display:block; margin:1em 0'>
1.E.3. If an individual Project Gutenberg&#8482; electronic work is posted
with the permission of the copyright holder, your use and distribution
must comply with both paragraphs 1.E.1 through 1.E.7 and any
additional terms imposed by the copyright holder. Additional terms
will be linked to the Project Gutenberg&#8482; License for all works
posted with the permission of the copyright holder found at the
beginning of this work.
</div>

<div style='display:block; margin:1em 0'>
1.E.4. Do not unlink or detach or remove the full Project Gutenberg&#8482;
License terms from this work, or any files containing a part of this
work or any other work associated with Project Gutenberg&#8482;.
</div>

<div style='display:block; margin:1em 0'>
1.E.5. Do not copy, display, perform, distribute or redistribute this
electronic work, or any part of this electronic work, without
prominently displaying the sentence set forth in paragraph 1.E.1 with
active links or immediate access to the full terms of the Project
Gutenberg&#8482; License.
</div>

<div style='display:block; margin:1em 0'>
1.E.6. You may convert to and distribute this work in any binary,
compressed, marked up, nonproprietary or proprietary form, including
any word processing or hypertext form. However, if you provide access
to or distribute copies of a Project Gutenberg&#8482; work in a format
other than &#8220;Plain Vanilla ASCII&#8221; or other format used in the official
version posted on the official Project Gutenberg&#8482; website
(www.gutenberg.org), you must, at no additional cost, fee or expense
to the user, provide a copy, a means of exporting a copy, or a means
of obtaining a copy upon request, of the work in its original &#8220;Plain
Vanilla ASCII&#8221; or other form. Any alternate format must include the
full Project Gutenberg&#8482; License as specified in paragraph 1.E.1.
</div>

<div style='display:block; margin:1em 0'>
1.E.7. Do not charge a fee for access to, viewing, displaying,
performing, copying or distributing any Project Gutenberg&#8482; works
unless you comply with paragraph 1.E.8 or 1.E.9.
</div>

<div style='display:block; margin:1em 0'>
1.E.8. You may charge a reasonable fee for copies of or providing
access to or distributing Project Gutenberg&#8482; electronic works
provided that:
</div>

<div style='margin-left:0.7em;'>
    <div style='text-indent:-0.7em'>
        &#8226; You pay a royalty fee of 20% of the gross profits you derive from
        the use of Project Gutenberg&#8482; works calculated using the method
        you already use to calculate your applicable taxes. The fee is owed
        to the owner of the Project Gutenberg&#8482; trademark, but he has
        agreed to donate royalties under this paragraph to the Project
        Gutenberg Literary Archive Foundation. Royalty payments must be paid
        within 60 days following each date on which you prepare (or are
        legally required to prepare) your periodic tax returns. Royalty
        payments should be clearly marked as such and sent to the Project
        Gutenberg Literary Archive Foundation at the address specified in
        Section 4, &#8220;Information about donations to the Project Gutenberg
        Literary Archive Foundation.&#8221;
    </div>

    <div style='text-indent:-0.7em'>
        &#8226; You provide a full refund of any money paid by a user who notifies
        you in writing (or by e-mail) within 30 days of receipt that s/he
        does not agree to the terms of the full Project Gutenberg&#8482;
        License. You must require such a user to return or destroy all
        copies of the works possessed in a physical medium and discontinue
        all use of and all access to other copies of Project Gutenberg&#8482;
        works.
    </div>

    <div style='text-indent:-0.7em'>
        &#8226; You provide, in accordance with paragraph 1.F.3, a full refund of
        any money paid for a work or a replacement copy, if a defect in the
        electronic work is discovered and reported to you within 90 days of
        receipt of the work.
    </div>

    <div style='text-indent:-0.7em'>
        &#8226; You comply with all other terms of this agreement for free
        distribution of Project Gutenberg&#8482; works.
    </div>
</div>

<div style='display:block; margin:1em 0'>
1.E.9. If you wish to charge a fee or distribute a Project
Gutenberg&#8482; electronic work or group of works on different terms than
are set forth in this agreement, you must obtain permission in writing
from the Project Gutenberg Literary Archive Foundation, the manager of
the Project Gutenberg&#8482; trademark. Contact the Foundation as set
forth in Section 3 below.
</div>

<div style='display:block; margin:1em 0'>
1.F.
</div>

<div style='display:block; margin:1em 0'>
1.F.1. Project Gutenberg volunteers and employees expend considerable
effort to identify, do copyright research on, transcribe and proofread
works not protected by U.S. copyright law in creating the Project
Gutenberg&#8482; collection. Despite these efforts, Project Gutenberg&#8482;
electronic works, and the medium on which they may be stored, may
contain &#8220;Defects,&#8221; such as, but not limited to, incomplete, inaccurate
or corrupt data, transcription errors, a copyright or other
intellectual property infringement, a defective or damaged disk or
other medium, a computer virus, or computer codes that damage or
cannot be read by your equipment.
</div>

<div style='display:block; margin:1em 0'>
1.F.2. LIMITED WARRANTY, DISCLAIMER OF DAMAGES - Except for the &#8220;Right
of Replacement or Refund&#8221; described in paragraph 1.F.3, the Project
Gutenberg Literary Archive Foundation, the owner of the Project
Gutenberg&#8482; trademark, and any other party distributing a Project
Gutenberg&#8482; electronic work under this agreement, disclaim all
liability to you for damages, costs and expenses, including legal
fees. YOU AGREE THAT YOU HAVE NO REMEDIES FOR NEGLIGENCE, STRICT
LIABILITY, BREACH OF WARRANTY OR BREACH OF CONTRACT EXCEPT THOSE
PROVIDED IN PARAGRAPH 1.F.3. YOU AGREE THAT THE FOUNDATION, THE
TRADEMARK OWNER, AND ANY DISTRIBUTOR UNDER THIS AGREEMENT WILL NOT BE
LIABLE TO YOU FOR ACTUAL, DIRECT, INDIRECT, CONSEQUENTIAL, PUNITIVE OR
INCIDENTAL DAMAGES EVEN IF YOU GIVE NOTICE OF THE POSSIBILITY OF SUCH
DAMAGE.
</div>

<div style='display:block; margin:1em 0'>
1.F.3. LIMITED RIGHT OF REPLACEMENT OR REFUND - If you discover a
defect in this electronic work within 90 days of receiving it, you can
receive a refund of the money (if any) you paid for it by sending a
written explanation to the person you received the work from. If you
received the work on a physical medium, you must return the medium
with your written explanation. The person or entity that provided you
with the defective work may elect to provide a replacement copy in
lieu of a refund. If you received the work electronically, the person
or entity providing it to you may choose to give you a second
opportunity to receive the work electronically in lieu of a refund. If
the second copy is also defective, you may demand a refund in writing
without further opportunities to fix the problem.
</div>

<div style='display:block; margin:1em 0'>
1.F.4. Except for the limited right of replacement or refund set forth
in paragraph 1.F.3, this work is provided to you &#8216;AS-IS&#8217;, WITH NO
OTHER WARRANTIES OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PURPOSE.
</div>

<div style='display:block; margin:1em 0'>
1.F.5. Some states do not allow disclaimers of certain implied
warranties or the exclusion or limitation of certain types of
damages. If any disclaimer or limitation set forth in this agreement
violates the law of the state applicable to this agreement, the
agreement shall be interpreted to make the maximum disclaimer or
limitation permitted by the applicable state law. The invalidity or
unenforceability of any provision of this agreement shall not void the
remaining provisions.
</div>

<div style='display:block; margin:1em 0'>
1.F.6. INDEMNITY - You agree to indemnify and hold the Foundation, the
trademark owner, any agent or employee of the Foundation, anyone
providing copies of Project Gutenberg&#8482; electronic works in
accordance with this agreement, and any volunteers associated with the
production, promotion and distribution of Project Gutenberg&#8482;
electronic works, harmless from all liability, costs and expenses,
including legal fees, that arise directly or indirectly from any of
the following which you do or cause to occur: (a) distribution of this
or any Project Gutenberg&#8482; work, (b) alteration, modification, or
additions or deletions to any Project Gutenberg&#8482; work, and (c) any
Defect you cause.
</div>

<div style='display:block; font-size:1.1em; margin:1em 0; font-weight:bold'>
Section 2. Information about the Mission of Project Gutenberg&#8482;
</div>

<div style='display:block; margin:1em 0'>
Project Gutenberg&#8482; is synonymous with the free distribution of
electronic works in formats readable by the widest variety of
computers including obsolete, old, middle-aged and new computers. It
exists because of the efforts of hundreds of volunteers and donations
from people in all walks of life.
</div>

<div style='display:block; margin:1em 0'>
Volunteers and financial support to provide volunteers with the
assistance they need are critical to reaching Project Gutenberg&#8482;&#8217;s
goals and ensuring that the Project Gutenberg&#8482; collection will
remain freely available for generations to come. In 2001, the Project
Gutenberg Literary Archive Foundation was created to provide a secure
and permanent future for Project Gutenberg&#8482; and future
generations. To learn more about the Project Gutenberg Literary
Archive Foundation and how your efforts and donations can help, see
Sections 3 and 4 and the Foundation information page at www.gutenberg.org.
</div>

<div style='display:block; font-size:1.1em; margin:1em 0; font-weight:bold'>
Section 3. Information about the Project Gutenberg Literary Archive Foundation
</div>

<div style='display:block; margin:1em 0'>
The Project Gutenberg Literary Archive Foundation is a non-profit
501(c)(3) educational corporation organized under the laws of the
state of Mississippi and granted tax exempt status by the Internal
Revenue Service. The Foundation&#8217;s EIN or federal tax identification
number is 64-6221541. Contributions to the Project Gutenberg Literary
Archive Foundation are tax deductible to the full extent permitted by
U.S. federal laws and your state&#8217;s laws.
</div>

<div style='display:block; margin:1em 0'>
The Foundation&#8217;s business office is located at 809 North 1500 West,
Salt Lake City, UT 84116, (801) 596-1887. Email contact links and up
to date contact information can be found at the Foundation&#8217;s website
and official page at www.gutenberg.org/contact
</div>

<div style='display:block; font-size:1.1em; margin:1em 0; font-weight:bold'>
Section 4. Information about Donations to the Project Gutenberg Literary Archive Foundation
</div>

<div style='display:block; margin:1em 0'>
Project Gutenberg&#8482; depends upon and cannot survive without widespread
public support and donations to carry out its mission of
increasing the number of public domain and licensed works that can be
freely distributed in machine-readable form accessible by the widest
array of equipment including outdated equipment. Many small donations
($1 to $5,000) are particularly important to maintaining tax exempt
status with the IRS.
</div>

<div style='display:block; margin:1em 0'>
The Foundation is committed to complying with the laws regulating
charities and charitable donations in all 50 states of the United
States. Compliance requirements are not uniform and it takes a
considerable effort, much paperwork and many fees to meet and keep up
with these requirements. We do not solicit donations in locations
where we have not received written confirmation of compliance. To SEND
DONATIONS or determine the status of compliance for any particular state
visit <a href="https://www.gutenberg.org/donate/">www.gutenberg.org/donate</a>.
</div>

<div style='display:block; margin:1em 0'>
While we cannot and do not solicit contributions from states where we
have not met the solicitation requirements, we know of no prohibition
against accepting unsolicited donations from donors in such states who
approach us with offers to donate.
</div>

<div style='display:block; margin:1em 0'>
International donations are gratefully accepted, but we cannot make
any statements concerning tax treatment of donations received from
outside the United States. U.S. laws alone swamp our small staff.
</div>

<div style='display:block; margin:1em 0'>
Please check the Project Gutenberg web pages for current donation
methods and addresses. Donations are accepted in a number of other
ways including checks, online payments and credit card donations. To
donate, please visit: www.gutenberg.org/donate
</div>

<div style='display:block; font-size:1.1em; margin:1em 0; font-weight:bold'>
Section 5. General Information About Project Gutenberg&#8482; electronic works
</div>

<div style='display:block; margin:1em 0'>
Professor Michael S. Hart was the originator of the Project
Gutenberg&#8482; concept of a library of electronic works that could be
freely shared with anyone. For forty years, he produced and
distributed Project Gutenberg&#8482; eBooks with only a loose network of
volunteer support.
</div>

<div style='display:block; margin:1em 0'>
Project Gutenberg&#8482; eBooks are often created from several printed
editions, all of which are confirmed as not protected by copyright in
the U.S. unless a copyright notice is included. Thus, we do not
necessarily keep eBooks in compliance with any particular paper
edition.
</div>

<div style='display:block; margin:1em 0'>
Most people start at our website which has the main PG search
facility: <a href="https://www.gutenberg.org">www.gutenberg.org</a>.
</div>

<div style='display:block; margin:1em 0'>
This website includes information about Project Gutenberg&#8482;,
including how to make donations to the Project Gutenberg Literary
Archive Foundation, how to help produce our new eBooks, and how to
subscribe to our email newsletter to hear about new eBooks.
</div>

</section>
'''
    return etree.fromstring(pg_footer, lxml.html.XHTMLParser())



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

AuxParser.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Open an url and return raw data.

"""


from ebookmaker.parsers import ParserBase

mediatypes = ('*/*', )

class Parser (ParserBase):
    """ Parse an auxiliary file. """
    auxparser = True
    def __init__ (self, attribs = None):
        ParserBase.__init__ (self, attribs)
        self.data = None


    def pre_parse (self):
        """ Parse the file. """
        self.data = self.bytes_content ()


    def serialize (self):
        """ Serialize file to string. """
        return self.data



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

HTMLParser.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""
import re
import unicodedata

from six.moves import urllib

import lxml.html
from lxml import etree

from bs4 import BeautifulSoup, NavigableString
from bs4.formatter import EntitySubstitution, HTMLFormatter


from libgutenberg.GutenbergGlobals import NS
from libgutenberg.Logger import critical, info, debug, warning, error
from libgutenberg.MediaTypes import mediatypes as mt

from ebookmaker import parsers
from ebookmaker.CommonCode import Options, EbookmakerBadFileException
from ebookmaker.utils import add_style, css_len, replace_elements, xpath
from . import HTMLParserBase
from .boilerplate import mark_soup

options = Options()

mediatypes = ('text/html', mt.xhtml)

RE_XMLDECL = re.compile(r'<\?xml[^?]+\?>\s*')

FONT_SIZES = {
    '1': 'xx-small',
    '2': 'x-small',
    '3': 'small',
    '4': 'medium',
    '5': 'large',
    '6': 'x-large',
    '7': 'xx-large',
    '+1': '110%',
    '-1': '90%',
    '+2': '125%',
    '-2': '75%',
}

LIST_STYLES = {
    '1': 'decimal',
    'i': 'lower-roman',
    'I': 'upper-roman',
    'a': 'lower-alpha',
    'A': 'upper-alpha',
}

REPLACE_ELEMENTS = {
    'applet': None,
    'script': None,
    'basefont': None,
    'object': 'div',
    'iframe': None,
    'isindex': None,
    'input': 'div',
    'legend': None,
    'fieldset': None,
    'font': 'span',
    'center': 'div',
    'blink': 'span',
    'embed': None,
}

DEPRECATED = {
    'align': 'hr', # should translate this to margin settings
    'alink': 'body',
    'background': 'body',
    'bgcolor': 'body',
    'border': 'img ',
    'compact': '*',
    'hspace': '*',
    'link': 'body',
    'noshade': 'hr',
    'nowrap': '*',
    'start': 'ol',
    'text': 'body',
    'value': 'li',
    'version': 'html',
    'vlink': 'body',
    'vspace': '*',
}

ALLOWED_IN_BODY = {
    NS.xhtml.address, NS.xhtml.article, NS.xhtml.blockquote, f'{NS.xhtml.de}l', NS.xhtml.div,
    NS.xhtml.dl, NS.xhtml.figure, NS.xhtml.footer, 
    NS.xhtml.h1, NS.xhtml.h2, NS.xhtml.h3, NS.xhtml.h4, NS.xhtml.h5, NS.xhtml.h6,
    NS.xhtml.header, NS.xhtml.hr, NS.xhtml.ins, NS.xhtml.noscript, NS.xhtml.ol, NS.xhtml.p,
    NS.xhtml.pre, NS.xhtml.section, NS.xhtml.script, NS.xhtml.table, NS.xhtml.ul,
    NS.svg.svg,
}

REPLACEMENTS = [
    ('*', 'bgcolor', 'color', lambda x: x),
    ('br', 'clear', 'clear', lambda x: x),
    ('caption div h1 h2 h3 h4 h5 h5 p', 'align', 'text-align', lambda x: x),
    ('hr', 'width', 'width', css_len),
    ('hr', 'size', 'border', css_len),
    ('img table', 'align', 'float', lambda x: x),
    ('font', 'color', 'color', lambda x: x),
    ('font', 'face', 'font-family', lambda x: x),
    ('font', 'size', 'font-size', lambda x: FONT_SIZES.get(x.strip(), 'medium')),
    ('table td th', 'height', 'height', css_len),
    ('table td th pre', 'width', 'width', css_len),
    ('li ol ul', 'type', 'list-style-type', lambda x: LIST_STYLES.get(x.strip(), '')),
]

CSS_FOR_REPLACED = {
    'blink': "",
    'center': "",
    'font': "",
}


def nfc(_str):
    return unicodedata.normalize('NFC', EntitySubstitution.substitute_xml(_str))

nfc_formatter = HTMLFormatter(entity_substitution=nfc)

class Parser(HTMLParserBase):
    """ Parse a HTML Text

    and convert it to xhtml suitable for ePub packaging.

    """

    @staticmethod
    def _fix_id(id_):
        """ Fix more common mistakes in ids.

        xml:id cannot start with digit, very common in pg.

        """

        if not parsers.RE_XML_NAME.match(id_):
            id_ = 'id_' + id_

        # debug("_fix_id: id = %s" % id_)
        return id_


    def _fix_internal_frag(self, id_):
        """ Fix more common mistakes in ids. """

        # Non-ascii hrefs must be %-escaped, but id attributes must
        # not be %-escaped.  Some producers %-escape ids to make id
        # and href look the same.  But '%' is an invalid character in
        # xml ids.
        #
        # See HTML 4.01 spec section B.2.

        if '%' in id_:
            try:
                bytes_id = urllib.parse.unquote(id_.encode('us-ascii')) # undo the %-escaping
                try:
                    id_ = bytes_id.decode('utf-8')
                except UnicodeError:
                    doc_encoding = self.attribs.orig_mediatype.params.get('charset')
                    if doc_encoding:
                        id_ = bytes_id.decode(doc_encoding)
            except Exception:
                pass # too broken to fix

        # An xml:id cannot start with a digit, a very common mistake
        # in pg.

        if not parsers.RE_XML_NAME.match(id_):
            id_ = 'id_' + id_

        if not parsers.RE_XML_NAME.match(id_):
            # still invalid ... we tried
            return None

        # debug("_fix_internal_frag: frag = %s" % id_)
        return id_


    def _fix_anchors(self):
        """ Move name to id and fix hrefs and ids. """

        def ids_and_names(xhtml):
            """iterator that runs over id attributes and name attributes"""
            for node in xpath(xhtml, "//xhtml:*[@id]"):
                yield node
            for node in xpath(xhtml, "//xhtml:a[@name]"):
                yield node

        # move anchor name to id
        # 'id' values are more strict than 'name' values
        # try to fix ill-formed ids
        if self.xhtml is None:
            return
        seen_ids = set()

        for anchor in ids_and_names(self.xhtml):
            id_ = anchor.get('id') or anchor.get('name')

            if 'name' in anchor.attrib:
                del anchor.attrib['name']
            if 'id' in anchor.attrib:
                del anchor.attrib['id']
            if NS.xml.id in anchor.attrib:
                del anchor.attrib[NS.xml.id]

            id_ = self._fix_id(id_)

            if not parsers.RE_XML_NAME.match(id_):
                error("Dropping ill-formed id '%s' in %s" % (id_, self.attribs.url))
                continue

            # well-formed id
            if id_ in seen_ids:
                error("Dropping duplicate id '%s' in %s" % (id_, self.attribs.url))
                continue

            seen_ids.add(id_)
            anchor.set('id', id_)


        # try to fix bogus fragment ids
        # 1. fragments point to xml:id, so must be well-formed ids
        # 2. the ids they point to must exist

        for link in xpath(self.xhtml, "//xhtml:*[@href]"):
            href = link.get('href')
            hre, frag = urllib.parse.urldefrag(href)
            if frag:
                frag = self._fix_internal_frag(frag)

                if not frag:
                    # non-recoverable ill-formed frag
                    del link.attrib['href']
                    self.add_class(link, 'pgkilled')
                    error('Dropping ill-formed frag in %s' % href)
                    continue

                # well-formed frag
                if hre:
                    # we have url + frag
                    link.set('href', "%s#%s" % (hre, urllib.parse.quote(frag.encode('utf-8'))))
                    self.add_class(link, 'pgexternal')
                elif frag in seen_ids:
                    # we have only frag
                    link.set('href', "#%s" % urllib.parse.quote(frag.encode('utf-8')))
                    self.add_class(link, 'pginternal')
                else:
                    del link.attrib['href']
                    self.add_class(link, 'pgkilled')
                    error("Dropping frag to non-existing id in %s" % href)


    def convert_epub_attribs(self):
        """ Parser leaves some data elements for HTML. Epubcheck doesn't like these.
        """
        for elem in xpath(self.xhtml, '//@epub:*/..'):
            for key in elem.attrib.keys():
                if key.startswith('{%s}' % NS.epub):
                    new_key = key.replace('{%s}' % NS.epub, 'data-epub-')
                    val = elem.attrib[key]
                    del elem.attrib[key]
                    elem.attrib[new_key] = val


    def enclose_text(self):
        """ same as setting enclose-text option on tidy;
        ' enclose any text it finds in the body element within a <P> element.
        This is useful when you want to take existing HTML and use it with a style sheet.'
        """
        for elem in self.xhtml.body:
            if elem.tag not in ALLOWED_IN_BODY:
                new_p = elem.makeelement(NS.xhtml.p)
                elem.addprevious(new_p)
                new_p.append(elem)

    def _to_xhtml11(self):
        """ Make vanilla xhtml more conform to xhtml 1.1 """

        # Change content-type meta to application/xhtml+xml.
        for meta in xpath(self.xhtml, "/xhtml:html/xhtml:head/xhtml:meta[@http-equiv]"):
            if meta.get('http-equiv').lower() == 'content-type':
                meta.getparent().remove(meta)

        # drop javascript

        for script in xpath(self.xhtml, "//xhtml:script"):
            script.drop_tree()

        # drop form

        for form in xpath(self.xhtml, "//xhtml:form"):
            form.drop_tree()

        # remove xml:space

        for elem in xpath(self.xhtml, "//xhtml:*[@xml:space]"):
            del elem.attrib[NS.xml.space]

        # blockquotes

        for bq in xpath(self.xhtml, "//xhtml:blockquote"):
            # no naked text allowed in <blockquote>
            div = etree.Element(NS.xhtml.div)
            for child in bq:
                div.append(child)
            div.text = bq.text
            bq.text = None
            bq.append(div)
            # lxml.html.defs.block_tags

        # insert tbody

        for table in xpath(self.xhtml, "//xhtml:table[xhtml:tr]"):
            # no naked <tr> allowed in <table>
            tbody = etree.Element(NS.xhtml.tbody)
            for tr in table:
                if tr.tag == NS.xhtml.tr:
                    tbody.append(tr)
            table.append(tbody)

        # wrap bare col elements

        for table in xpath(self.xhtml, "//xhtml:table[xhtml:col]"):
            # epub3 only allows col elements inside colgroup elements
            colgroup = etree.Element(NS.xhtml.colgroup)
            for col in table:
                if col.tag == NS.xhtml.col:
                    colgroup.append(col)
            table.insert(0, colgroup)

        # move lang to xml:lang

        for elem in xpath(self.xhtml, "//xhtml:*[@lang]"):
            # bug in lxml 2.2.2: sometimes deletes wrong element
            # so we delete both and reset the right one
            lang = elem.get('lang')
            try:
                del elem.attrib[NS.xml.lang]
            except KeyError:
                pass
            del elem.attrib['lang']
            elem.set(NS.xml.lang, lang)

        # strip deprecated attributes
        for (tags, attr, cssattr, val2css) in REPLACEMENTS:
            for tag in tags.split():
                for elem in xpath(self.xhtml, f"//xhtml:{tag}[@{attr}]"):
                    if elem.attrib[attr]:
                        val = elem.attrib[attr]
                    del elem.attrib[attr]
                    if cssattr:
                        add_style(elem, style=f'{cssattr}: {val2css(val)};')

        # if source has epub attributes, change them to data attributes
        self.convert_epub_attribs()

        for a, t in DEPRECATED.items():
            for tag in t.split():
                for elem in xpath(self.xhtml, "//xhtml:%s[@%s]" % (tag, a)):
                    del elem.attrib[a]

        # strip empty class attributes

        for elem in xpath(self.xhtml,
                          "//xhtml:*[@class and normalize-space(@class) = '']"):
            del elem.attrib['class']

        # fix attribute values
        attrs_to_fix = [('align frame rules', lambda x: x.lower())]
        for attrs, fix in attrs_to_fix:
            for attr in attrs.split():
                for elem in xpath(self.xhtml, f"//xhtml:*[@{attr}]"):
                    elem.attrib[attr] = fix(elem.attrib[attr])

        # strip bogus header markup by Joe L.
        for elem in xpath(self.xhtml, "//xhtml:h1"):
            if elem.text and elem.text.startswith("The Project Gutenberg eBook"):
                elem.tag = NS.xhtml.p
        for elem in xpath(self.xhtml, "//xhtml:h3"):
            if elem.text and elem.text.startswith("E-text prepared by"):
                elem.tag = NS.xhtml.p

        # deprecated elements -  replace with <span/div class="xhtml_{tag name}">
        deprecated_used = replace_elements(self.xhtml, REPLACE_ELEMENTS)

        # enclose text the way tidy does
        self.enclose_text()

        ##### cleanup #######

        css_for_deprecated = ' '.join([CSS_FOR_REPLACED.get(tag, '') for tag in deprecated_used])
        if css_for_deprecated.strip():
            print(f'css_for_deprecated: {css_for_deprecated}')
            elem = etree.Element(NS.xhtml.style)
            elem.text = css_for_deprecated
            self.xhtml.find(NS.xhtml.head).insert(1, elem) # right after charset declaration


    def _make_coverpage_link(self, coverpage_url=None):
        """ Insert a <link rel="icon"> in the html head.

        First we determine the coverpage url.  In HTML we find the
        coverpage by appling these rules:

          0. the image specified by the --cover command-line option
          1. the image specified in <link rel='icon'> or <link rel='coverpage'>,
          2. the image with an id of 'coverpage' or
          3. the image with an url containing 'cover'
          4. the image with an url containing 'title'

        If one rule returns images we take the first one in document
        order, else we proceed with the next rule.
        """

        coverpages = xpath(self.xhtml, "//xhtml:link[@rel='icon' or @rel='coverpage']")
        cover_attrs = {'rel': 'icon', 'type': 'image/x-cover'}
        if coverpage_url:
            for coverpage in coverpages:
                coverpage.set('href', coverpage_url)
                coverpage.attrib.update(cover_attrs)
                debug("overrode link to coverpage with %s." % coverpage_url)

            else:
                for head in xpath(self.xhtml, "/xhtml:html/xhtml:head"):
                    head.append(parsers.em.link(rel='icon', href=coverpage_url,
                                                type='image/x-cover'))
                    debug("Inserted link to coverpage %s." % coverpage_url)
            return

        for coverpage in coverpages:
            if 'type' in coverpage.attrib and coverpage.attrib['type'] != 'image/x-cover':
                continue
            url = coverpage.get('href')
            debug("Found link to coverpage %s." % url)
            if coverpage.attrib['rel'] == 'coverpage':
                coverpage.attrib.update(cover_attrs)
            return   # already provided by user

        # look for a suitable candidate
        coverpages = xpath(self.xhtml, "//xhtml:img[@id='coverpage']")
        if not coverpages:
            coverpages = xpath(self.xhtml, "//xhtml:img[contains(@src, 'cover')]")
        if not coverpages:
            coverpages = xpath(self.xhtml, "//xhtml:img[contains(@src, 'title')]")

        for coverpage in coverpages:
            for head in xpath(self.xhtml, "/xhtml:html/xhtml:head"):
                url = coverpage.get('src')
                head.append(parsers.em.link(rel='icon', href=url, type='image/x-cover'))
                debug("Inserted link to coverpage %s." % url)


    def __parse(self, html):
        def fix_bad_entity(entity, html):
            warning("Fixing problem entity: '%s'" % entity)
            soup = BeautifulSoup(f'&{entity};', 'html.parser')
            if soup:
                should_be = str(soup)
                return html.replace(f'&{entity};', should_be)
            return html.replace(f'&{entity};', entity)
        # remove xml decl and doctype, we will add the correct one before serializing
        # html = re.compile('^.*<html ', re.I | re.S).sub('<html ', html)

        re_xml_decl = re.compile(r'^.*?<\?xml.*?\?>', re.S)
        re_doctype = re.compile(r'<!DOCTYPE[^>]*>\s*', re.I)
        html = re_xml_decl.sub('', html)
        html = re_doctype.sub('<!DOCTYPE html >', html)

        try:
            return etree.fromstring(
                html,
                lxml.html.XHTMLParser(huge_tree=True),
                base_url=self.attribs.url)
        except etree.ParseError as what:
            # cannot try HTML parser because we depend on correct xhtml namespace
            m = re.search(r"Entity '([^']+)'", str(what))
            if m and m.group(1):
                html = fix_bad_entity(m.group(1), html)
                return self.__parse(html)
            else:
                error("Failed to parse file because: %s" % what)
                mline = re.search(r'line\s(\d+),', str(what))
                if mline:
                    lineno = int(mline.group(1))
                    if html:
                        error("Line %d: %s" % (lineno, html.splitlines()[lineno - 1]))
                    else:
                        error("empty document")


    def pre_parse(self):
        """
        Pre-parse a html ebook.

        Does a full parse because a lightweight parse would be almost
        as much work.

        """

        # cache
        if self.xhtml is not None:
            return

        debug("HTMLParser.pre_parse() ...")

        try:
            soup = BeautifulSoup(self.bytes_content(), 'lxml', exclude_encodings=["us-ascii"])
        except:
            critical('failed to parse %s', self.attribs.url)
            raise EbookmakerBadFileException('failed parsing')
        try:
            if not soup.html.body.contents:
                critical('body has no contents in ', self.attribs.url)
                raise EbookmakerBadFileException('body has no contents')
        except:
            critical('%s is not a usable html file', self.attribs.url)
            raise EbookmakerBadFileException('unusable file')

        soup.html['xmlns'] = NS.xhtml

        # ancient browsers didn't understand stylesheets, so xml comments were used to hide the
        # style text. Our CSS parser is too modern to remember this, it seems. 
        # So we needed to un-comment the style text
        xmlcomment = re.compile(r'<!--(.*?)-->', re.S)
        for commented_style in soup.find_all('style', string=xmlcomment):
            commented_style.string = xmlcomment.sub(r'\1', str(commented_style.string))

        xmlopencomment = re.compile(r'<!--(.*?)')
        for commented_style in soup.find_all('style', string=xmlopencomment):
            commented_style.string = xmlopencomment.sub(r'', str(commented_style.string))

        # pg producers did crazy things in css comments, such as inserting CDATA sections
        csscomment = re.compile(r'/\*(.*?)\*/', re.S)
        for commented_style in soup.find_all('style', string=csscomment):
            commented_style.string = csscomment.sub('', str(commented_style.string))

        # wrap bare strings at body top level
        for elem in soup.html.body.contents:
            if isinstance(elem, NavigableString) and str(elem).strip(' \n\r\t'):
                p = soup.new_tag('p')
                p.string = str(elem)
                elem.replace_with(p)

        marked = mark_soup(soup)
        if not marked:
            warning('no boilerplate found in %s', self.attribs.url)

        html = soup.decode(formatter=nfc_formatter)
        if not html:
            critical('no content in %s', self.attribs.url)
            return

        if '\r' in html or '\u2028' in html:
            html = '\n'.join(html.splitlines())
        self.unicode_buffer = html

        self.xhtml = self.__parse(html)     # let exception bubble up

        self._fix_anchors() # needs relative paths

        self.xhtml.make_links_absolute(base_url=self.attribs.url)

        self._to_xhtml11()
        self._make_coverpage_link()

        debug("Done parsing %s", self.attribs.url)


    def parse(self):
        """ Fully parse a html ebook. """

        debug("HTMLParser.parse() ...")

        self.pre_parse()



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

WriterFactory.py

Copyright 2009-14 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Writer factory. Dynamically loads writers from directories.

"""


import os.path

from pkg_resources import resource_listdir # pylint: disable=E0611

from libgutenberg.Logger import error, debug
from ebookmaker.CommonCode import Options

options = Options()

writers = {}

def __load_writers_from (package_name):
    """ See what types we can write. """

    for fn in resource_listdir (package_name, ''):
        modulename, ext = os.path.splitext (fn)
        if ext == '.py' and modulename.endswith ('Writer'):
            type_ = modulename.lower ().replace ('writer', '')
            try:
                debug ("Loading writer type %s from module %s" % (type_, modulename))
                module = __import__ (package_name + '.' + modulename, fromlist = [modulename])
                writers[type_] = module
            except ImportError as what:
                error (
                    "Could not load writer type %s from module %s. %s" %
                    (type_, modulename, what)
                )


def load_writers ():
    """ See what types we can write. """

    __load_writers_from ('ebookmaker.writers')

    for package in options.extension_packages:
        __load_writers_from (package)

    return writers.keys ()


def unload_writers ():
    """ Unload writer modules. """
    for k in writers.keys ():
        del writers[k]


def create (type_):
    """ Load writer module for type. """
    try:
        if type_ == 'kf8':
            type_ = 'kindle'
        return writers[type_].Writer ()
    except KeyError:
        raise KeyError ('No writer for type %s' % type_)



#
# ebookmaker distribution
#
#!/usr/bin/env python

from setuptools import setup

VERSION = '0.12.13'

if __name__ == "__main__":
 
    setup (
        name = 'ebookmaker',
        version = VERSION,

        packages = [
            'ebookmaker',
            'ebookmaker.parsers',
            'ebookmaker.writers',
            'ebookmaker.packagers',

            'ebookmaker.mydocutils',
            'ebookmaker.mydocutils.parsers',
            'ebookmaker.mydocutils.transforms',
            'ebookmaker.mydocutils.writers',

            'ebookmaker.mydocutils.gutenberg',
            'ebookmaker.mydocutils.gutenberg.parsers',
            'ebookmaker.mydocutils.gutenberg.transforms',
            'ebookmaker.mydocutils.gutenberg.writers',
        ],

        scripts = [
            'scripts/ebookmaker',
            'scripts/convert_unitame',
            'scripts/rhyme_compiler',
        ],

        install_requires = [
            'pillow>=8.3.2',
            'cherrypy',
            'cssutils',
            'docutils>=0.18.1',
            'lxml',
            'roman',
            'requests',
            'six>=1.4.1',
            'libgutenberg[covers]>=0.10.0',
            'cchardet',
            'beautifulsoup4',
        ],
    
        package_data = {
            'ebookmaker.parsers': ['broken.png', 'txt2all.css'],
            'ebookmaker.writers': ['cover.jpg'],
            'ebookmaker.mydocutils.parsers': ['*.rst'],
            'ebookmaker.mydocutils.writers': ['*.css'],
            'ebookmaker.mydocutils.gutenberg.parsers': ['*.rst'],
        },

        data_files = [
            ('', ['CHANGES', 'README.md']),
        ],

        # metadata for upload to PyPI

        author = "Marcello Perathoner",
        maintainer = "Eric Hellman",
        maintainer_email = "eric@hellman.net",
        description = "The Project Gutenberg tool to generate EPUBs and other ebook formats.",
        long_description = open ('README.md', encoding='utf-8').read (),
        long_description_content_type = 'text/markdown',
        license = "GPL v3",
        keywords = "ebook epub kindle pdf rst reST reStructuredText project gutenberg format conversion",
        url = "https://github.com/gutenbergtools/ebookmaker/",

        classifiers = [
            "Topic :: Text Processing",
            "License :: OSI Approved :: GNU General Public License (GPL)",
            "Environment :: Console",
            "Operating System :: OS Independent",
            "Intended Audience :: Other Audience",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],

        platforms = 'OS-independent'
    )



# -*- coding: utf-8 -*-
#
# $Id: manpage.py 6270 2010-03-18 22:32:09Z milde $
# Author: Engelbert Gruber <grubert@users.sourceforge.net>
# Copyright: This module is put into the public domain.
#
# Rewritten almost completely
# by Marcello Perathoner <marcello@perathoner.de>

"""

Nroff writer for reStructuredText. Tweaked for Project Gutenberg usage.

"""

from __future__ import unicode_literals

__docformat__ = 'reStructuredText'

from ebookmaker.mydocutils.writers import nroff
from ebookmaker import Unitame

from libgutenberg.Logger import info, debug, warning, error

GUTENBERG_NROFF_PREAMBLE = r""".\" -*- mode: nroff -*- coding: {encoding} -*-
.\" This file produces Project Gutenberg plain text. Usage:
.\"   $ groff -t -K {device} -T {device} this_file > output.txt
.
.pl 100000       \" very tall page: disable pagebreaks
.ll 72m
.po 0
.ad l           \" text-align: left
.nh             \" hyphenation: off
.cflags 0 .?!   \" single sentence space
.cflags 0 -\[hy]\[em]   \" don't break on -
.
.de nop
..
.blm nop        \" do nothing on empty line
.
.nr [env_cnt] 0
.ev 0           \" start in a defined environment
.
.de push_env
.br
.nr last_env \\n[.ev]            \" save current environment name
.nr env_cnt +1   \" generate new environment name
.ev \\n[env_cnt]
.evc \\n[last_env]
..
.de pop_env
.br
.ev
.nr env_cnt -1
..
.
"""

GUTENBERG_NROFF_POSTAMBLE = r""".
.pl 0    \" ends very long page here
.\" End of File
"""

class Writer (nroff.Writer):
    """ A plaintext writer thru nroff. """

    supported = ('pg-nroff',)
    """Formats this writer supports."""

    def __init__ (self):
        nroff.Writer.__init__ (self)
        self.translator_class = Translator

    def translate (self):
        visitor = self.translator_class (self.document)
        del Unitame.unhandled_chars[:]
        self.document.walkabout (visitor)
        self.output = visitor.astext ()
        if Unitame.unhandled_chars:
            error ("unitame: unhandled chars: %s" % ", ".join (set (Unitame.unhandled_chars)))

    #def get_transforms (self):
    #    tfs = writers.Writer.get_transforms (self)
    #    return tfs + [parts.CharsetTransform]



class Translator (nroff.Translator):
    """ nroff translator """

    def preamble (self):
        """ Inserts nroff preamble. """
        return GUTENBERG_NROFF_PREAMBLE.format (
            encoding = self.encoding, device = self.device)


    def postamble (self):
        """ Inserts nroff postamble. """
        return GUTENBERG_NROFF_POSTAMBLE.format (
            encoding = self.encoding, device = self.device)


    def init_translate_maps (self):
        nroff.Translator.init_translate_maps (self)

        update = {
            0x0011: r"\~",       # nbsp, see: Unitame.py
            0x0012: r"\%",       # shy,  see: Unitame.py
            }

        self.translate_map.update (update)
        self.translate_map_literal.update (update)


    def register_classes (self):
        """ Register classes.

        Use the idiosyncratic PG convention of marking up italics etc.

        """

        #
        # This does not call the base class !!!
        #

        self.register_class ('simple', 'left',         '.ad l', '')
        self.register_class ('simple', 'right',        '.ad r', '')
        self.register_class ('simple', 'center',       '.ad c', '')

        self.register_class ('inline', 'italics',      '_',    '_')
        self.register_class ('inline', 'bold',         '*',    '*')

        self.register_class ('inline', 'monospaced',   '',     '')
        self.register_class ('inline', 'superscript',  '',     '')
        self.register_class ('inline', 'subscript',    '',     '')

        self.register_class ('inline', 'small-caps',   '_',    '_')
        self.register_class ('inline', 'gesperrt',     '_',    '_')
        self.register_class ('inline', 'antiqua',      '_',    '_')
        self.register_class ('inline', 'larger',       '',     '')
        self.register_class ('inline', 'smaller',      '',     '')


    def translate (self, text):
        """ Reduce the charset while keeping text a unicode string. """

        # NOTE: there's an alternate approach in
        # transforms.parts.CharsetTransform

        if self.encoding != 'utf-8':
            text = text.encode (self.encoding, 'unitame')
            text = text.decode (self.encoding)

        if self.in_literal:
            text = text.translate (self.translate_map_literal)
        else:
            text = text.translate (self.translate_map)

        return text


    def visit_inner (self, node):
        """ Try to remove duplicated PG highlight markers. """
        if node.type == 'inline':
            prefixes = self.get_prefix (node.type, node['classes'])
            for prefix in prefixes:
                if prefix == self.last_output_char:
                    self.backspace ()
                else:
                    self.text (prefix)
        else:
            nroff.Translator.visit_inner (self, node)


    def visit_inline (self, node):
        if 'toc-pageref' in node['classes']:
            maxlen = 3 # sensible default
            while node.parent:
                node = node.parent
                if 'pageno_maxlen' in node:
                    maxlen = node['pageno_maxlen']
                    break
            self.cmd (('linetabs 1',
                       r'ta (\n[.l]u - \n[.i]u - %dm) +%dmR' % (maxlen + 1, maxlen + 1),
                       r'lc .'))
            self.text (chr (1) + '\t')
        nroff.Translator.visit_inline (self, node)

    def visit_section_title (self, node):
        """ Implements PG-standard spacing before headers. """
        self.sp (max (2, 5 - self.section_level))

    def visit_figure (self, node):
        self.sp (1)
        self.push ()

    def depart_figure (self, node):
        self.pop ()
        self.sp (1)

    def visit_image (self, node):
        # ignore alt attribute except for dropcaps
        if 'dropcap' in node['classes']:
            self.text (node.attributes.get ('alt', ''))

    def visit_page (self, node):
        if 'clearpage' in node['classes']:
            self.sp (4)
        elif 'cleardoublepage' in node['classes']:
            self.sp (4)
        else:
            nroff.Translator.visit_page (self, node)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
PDFWriter.py

Copyright 2011 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Convert RST to PDF.

"""


import os
import subprocess

from libgutenberg.Logger import debug, info, warning, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat, mkdir_for_filename

from ebookmaker import ParserFactory
from ebookmaker import writers
from ebookmaker.CommonCode import Options

options = Options()

class Writer (writers.BaseWriter):
    """ Class to write PDF. """

    def build (self, job):
        """ Build PDF file. """

        inputfilename  = job.url
        outputfilename = os.path.join (os.path.abspath(job.outputdir), job.outputfile)

        debug ("Inputfile: %s" % inputfilename)
        info ("Creating PDF file: %s" % outputfilename)

        mkdir_for_filename(outputfilename)
        parser = ParserFactory.ParserFactory.create (inputfilename)

        if not hasattr (parser, 'rst2xetex'):
            warning ('Skipping PDF Output because input mediatype is %s' % parser.mediatype())
            raise SkipOutputFormat

        # Brain-dead xetex doesn't understand unix pipes
        # so we have to write a temp file

        texfilename = os.path.splitext (outputfilename)[0] + '.tex'
        auxfilename = os.path.splitext (outputfilename)[0] + '.aux'
        logfilename = os.path.splitext (outputfilename)[0] + '.log'

        try:
            os.remove (auxfilename)
        except OSError:
            pass

        tex = parser.rst2xetex (job)
        with open (texfilename, 'wb') as fp:
            fp.write (tex)

        try:
            cwd = os.getcwd ()
            os.chdir (os.path.abspath(job.outputdir))

            _xetex = subprocess.Popen ([options.config.XELATEX,
                                        "-output-directory", job.outputdir,
                                        "-interaction", "nonstopmode",
                                        texfilename],
                                       stdin = subprocess.PIPE,
                                       stdout = subprocess.PIPE,
                                       stderr = subprocess.PIPE)
        except OSError as what:
            os.chdir (cwd)
            error ("PDFWriter: %s %s" % (options.config.XELATEX, what))
            raise SkipOutputFormat

        (dummy_stdout, dummy_stderr) = _xetex.communicate ()

        with open (logfilename, encoding='utf-8') as fp:
            for line in fp:
                line = line.strip ()
                if 'Error:' in line:
                    error ("xetex: %s" % line)
                if options.verbose >= 1:
                    if 'Warning:' in line:
                        warning ("xetex: %s" % line)

        if options.verbose < 2:
            try:
                os.remove (texfilename)
                os.remove (logfilename)
                os.remove (auxfilename)
            except OSError:
                pass

        os.chdir (cwd)

        info ("Done PDF file: %s" % outputfilename)



#!/usr/bin/env python


"""
CommonCode.py

Copyright 2014-2021 by Marcello Perathoner and Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Common code for EbookMaker and EbookConverter.

"""
import os
import os.path

from six.moves import configparser

from libgutenberg.CommonOptions import Options

class Struct(object):
    pass

options = Options()

class EbookmakerBadFileException(Exception):
    pass

class Job(object):
    """Hold 'globals' for a job.

    A job is defined as one unit of work, acting on one input url.

    """

    def __init__(self, type_):
        self.type = type_
        self.maintype, self.subtype = os.path.splitext(self.type)

        self.url = None
        self.outputdir = None
        self.outputfile = None
        self.logfile = None
        self.dc = None
        self.source = None
        self.opf_identifier = None
        self.main = None
        self.link_map = {}


    def __str__(self):
        l = []
        for k, v in self.__dict__.items():
            l.append("%s: %s" % (k, v))
        return '\n'.join(l)


def add_dependencies(targets, deps, order=None):
    """ Add dependent formats and optionally put into right build order. """

    for target, deps in deps.items():
        if target in targets:
            targets = list(set(targets).union(deps))
    if order:
        return list(filter(targets.__contains__, order))
    return targets


def add_common_options(ap, user_config_file):
    """ Add aptions common to all programs. """

    ap.add_argument(
        "--verbose", "-v",
        action="count",
        default=0,
        help="be verbose (-v -v be more verbose)")

    ap.add_argument(
        "--config",
        metavar="CONFIG_FILE",
        dest="config_file",
        action="store",
        default=user_config_file,
        help="read config file (default: %(default)s)")

    ap.add_argument(
        "--validate",
        dest="validate",
        action="store_true",
        help="validate epub and html through epubcheck/nu")

    ap.add_argument(
        "--notify",
        dest="notify",
        action="store_true",
        help="write CRITICAL messages to notifier logs")


def set_arg_defaults(ap, config_file):
    # get default command-line args
    cp = configparser.ConfigParser()
    cp.read(config_file)
    if cp.has_section('DEFAULT_ARGS'):
        ap.set_defaults(**dict(cp.items('DEFAULT_ARGS')))

def parse_config_and_args(ap, sys_config, defaults=None):

    # put command-line args into options
    options.update(vars(ap.parse_args()))

    cp = configparser.ConfigParser()
    cp.read((sys_config, options.config_file))

    options.config = Struct()

    for name, value in defaults.items():
        setattr(options.config, name.upper(), value)

    for section in cp.sections():
        for name, value in cp.items(section):
            setattr(options.config, name.upper(), value)

    return options


PRIVATE = os.getenv('PRIVATE') or ''
NOTIFICATION_DIR = os.path.join(PRIVATE, 'logs', 'notifications')

def queue_notifications(ebook, message, subject='EbookMaker Notification'):
    message_queue = os.path.join(NOTIFICATION_DIR, '%s.txt' % ebook)
    with open(message_queue, 'a+') as messagefile:
        messagefile.write('Subject: %s\n' % subject)
        messagefile.write(message)


def dir_from_url(url):
    if os.path.isdir(url):
        return url
    if url.startswith('file://'):
        dir = os.path.dirname(os.path.abspath(url[7:]))
    elif url.startswith('file:'):
        dir = os.path.dirname(os.path.abspath(url[5:]))
    else:
        dir = os.path.dirname(os.path.abspath(url))
    return dir


def find_candidates(path, file_filter=lambda x: True):
    """ walk the directory containing path, return files satisfying file_filter 
    """
    path = dir_from_url(path)
    for (root, dirs, files) in os.walk(path):
        if '/.' in root or root.startswith('.'):
            continue
        for fname in files:
            fpath = os.path.join(root, fname)
            if file_filter(fpath):
                yield fpath
